/* file: "x11.c" */

/*--------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>

typedef unsigned int U32;
typedef U32 RGB;

#define RGBCOLOR(r,g,b)(((((U32)(r)<<8)+(U32)(g))<<8)+(U32)(b))
#define RED(c)((c)>>16)
#define GREEN(c)(((c)>>8)&0xff)
#define BLUE(c)((c)&0xff)

#define WHITE 0
#define BLACK 1
#define GRAY  2
#define OTHER 3

#define CENTER 0
#define LEFT   1
#define RIGHT  2

#define MAX_NB_COLORS (256+4)
#define INVERT (MAX_NB_COLORS-1)

typedef struct {
  char *name;
  int w, h;
  Window win;
  XFontStruct *font;
  GC col[MAX_NB_COLORS];
  RGB *colors;
  int nbcolors;
  } window_struct;

typedef window_struct *window;

int graph_display = 1;

int graph_black_and_white = 0;

int graph_screen_width = 1280, graph_screen_height = 1024;

RGB rgb_white = RGBCOLOR(255,255,255);  /* white */
RGB rgb_black = RGBCOLOR(  0,  0,  0);  /* black */
RGB rgb_gray  = RGBCOLOR(191,191,191);  /* light gray */

RGB win_get_color (win, col)
window win;
int col;
{
  if (col == WHITE)
    return rgb_white;
  else if (col == BLACK)
    return rgb_black;
  else if (col == GRAY)
    return rgb_gray;
  else
    return win->colors[col-OTHER];
}

#define STIPPLE_W 8
#define STIPPLE_H 8

unsigned char stipple[][STIPPLE_H] = {
{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff },
{ 0xaa, 0x55, 0xaa, 0x55, 0xaa, 0x55, 0xaa, 0x55 },
{ 0xaa, 0xaa, 0xaa, 0xaa, 0xaa, 0xaa, 0xaa, 0xaa },
{ 0x00, 0x80, 0x00, 0x08, 0x00, 0x80, 0x00, 0x08 },
{ 0x11, 0x22, 0x44, 0x88, 0x11, 0x22, 0x44, 0x88 },
{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff },
{ 0x88, 0x22, 0x88, 0x22, 0x88, 0x22, 0x88, 0x22 },
{ 0x00, 0xaa, 0x00, 0xaa, 0x00, 0xaa, 0x00, 0xaa },
{ 0xaa, 0x55, 0xaa, 0x55, 0xaa, 0x55, 0xaa, 0x55 },
{ 0xee, 0xdd, 0xbb, 0x77, 0xee, 0xdd, 0xbb, 0x77 },
{ 0xdd, 0x77, 0xdd, 0x77, 0xdd, 0x77, 0xdd, 0x77 },
{ 0xdd, 0xff, 0x77, 0xff, 0xdd, 0xff, 0x77, 0xff },
{ 0x00, 0xff, 0x00, 0xff, 0x00, 0xff, 0x00, 0xff },
{ 0x80, 0x10, 0x02, 0x20, 0x01, 0x08, 0x40, 0x04 },
{ 0x88, 0xff, 0x88, 0x88, 0x88, 0xff, 0x88, 0x88 },
{ 0xb1, 0x30, 0x03, 0x1b, 0xd8, 0xc0, 0x0c, 0x8d },
{ 0x00, 0x80, 0x40, 0x20, 0x00, 0x02, 0x04, 0x08 }
};

unsigned char *win_get_stipple (win, col)
window win;
int col;
{
  if (col < OTHER)
    return stipple[col];
  else
    return stipple[OTHER+(col-OTHER)%(sizeof(stipple)/STIPPLE_H-OTHER)];
}

#define NB_RGB_COLORS 23

RGB rgb_colors[] =
{ RGBCOLOR(255,255,255)  /* white          */
, RGBCOLOR(  0,  0,  0)  /* black          */
, RGBCOLOR(253,188,175)  /* pink           */
, RGBCOLOR(253,177,165)  /* pink           */
, RGBCOLOR(249,157,144)  /* pink           */
, RGBCOLOR(191, 50, 36)  /* pink           */
, RGBCOLOR(192,249,126)  /* green          */
, RGBCOLOR(185,253,123)  /* green          */
, RGBCOLOR(167,236,108)  /* green          */
, RGBCOLOR( 82,166, 28)  /* green          */
, RGBCOLOR(148,214, 87)  /* green          */
, RGBCOLOR(112,144,202)  /* LightSlateBlue */
, RGBCOLOR(255,255,  0)  /* yellow         */
, RGBCOLOR(255,  0,255)  /* magenta        */
, RGBCOLOR(  0,255,255)  /* cyan           */
, RGBCOLOR(255,215,  0)  /* gold           */
, RGBCOLOR(255,  0,  0)  /* red            */
, RGBCOLOR(152,251,152)  /* PaleGreen      */
, RGBCOLOR(160, 32,240)  /* purple         */
, RGBCOLOR(190,190,190)  /* gray           */
, RGBCOLOR(165, 42,  0)  /* brown          */
, RGBCOLOR(255,248,220)  /* cornsilk       */
, RGBCOLOR(255,255,240)  /* ivory          */
};

/*--------------------------------------------------------------------------*/

Display *display;
int screen;
window win;


void g_flush()
{
  XFlush( display );
}


void g_clear()
{ XClearWindow( display, win->win );
}


void g_box(color, x, y, w, h )
int color;
int x, y, w, h;
{
  //  printf ("color=%d x=%d y=%d\n", color, x, y);
  XFillRectangle( display, win->win, win->col[color], x, win->h - (y+h), w, h );
}


void g_text( color, where, x, y, str )
int color;
int where;
int x, y;
char *str;
{ int len, str_width, y_pos, x_pos;

//  printf ("color=%d x=%d y=%d\n", color, x, y);

  len = strlen( str );
  str_width = XTextWidth( win->font, str, len );
  y_pos = win->h - (y - win->font->max_bounds.ascent/2);

  switch (where)
  { case CENTER: x_pos = x - str_width/2; break;
    case LEFT:   x_pos = x;               break;
    case RIGHT:  x_pos = x - str_width;   break;
  }

  XDrawString( display, win->win, win->col[color], x_pos, y_pos, str, len );
}


void g_init (name, width, height)
char *name;
int width, height;
{
  int         i;
  XGCValues   values;
  Colormap    cmap;
  int         win_x, win_y, win_w, win_h, win_b;
  XEvent      report;

  /* connect to X server */

  if ( (display=XOpenDisplay(NULL)) == NULL )
    {
      fprintf( stderr, "can't connect to X server\n");
      exit (1);
    }

  /* get screen size */

  screen = DefaultScreen( display );

  graph_screen_width  = DisplayWidth( display, screen );
  graph_screen_height = DisplayHeight( display, screen );

  win = (window)malloc( sizeof(*win) );
  if (win == NULL)
  { fprintf( stderr, "can't allocate window\n" );
    exit (1);
  }

  win->name    = name;
  win->w       = width;
  win->h       = height;
  win->colors  = rgb_colors;
  win->nbcolors= NB_RGB_COLORS;

  /* get color map */

  cmap = DefaultColormap( display, screen );

  /* create a window */

  win_x = 0;      win_y = 0;       /* position */
  win_w = win->w; win_h = win->h;  /* size     */
  win_b = 2;                       /* border   */

  win->win = XCreateSimpleWindow( display,
                                  RootWindow( display, screen ),
                                  win_x, win_y,
                                  win_w, win_h,
                                  win_b, BlackPixel( display, screen ),
                                  WhitePixel( display, screen ) );

  /* set graphic context */

  if ((win->font=XLoadQueryFont(display,"5x8")) == NULL)
    if ((win->font=XLoadQueryFont(display,"fixed")) == NULL)
    { fprintf( stderr, "can't open 6x10 or fixed font\n");
      exit (1);
    }

  for (i=0; i<win->nbcolors+OTHER; i++)
  { if (graph_black_and_white || (DisplayPlanes( display, screen ) == 1))
    { values.stipple = XCreateBitmapFromData( display,
                                              RootWindow( display, screen ),
                                              win_get_stipple(win, i), STIPPLE_W, STIPPLE_H);
      values.fill_style = FillOpaqueStippled;
      values.foreground = BlackPixel( display, screen );
      values.background = WhitePixel( display, screen );
    }
    else
    { XColor def;
      static char colorname[] = "#XXXXXX";
      static char hex[] = "0123456789abcdef";
      RGB rgb = win_get_color(win, i);
      int j, c = RED(rgb)*65536+GREEN(rgb)*256+BLUE(rgb);

      for (j=6; j>0; j--) { colorname[j] = hex[c%16]; c /= 16; }

      if (!XParseColor( display, cmap, colorname, &def ))
      { fprintf( stderr, "color name %s not in database\n", colorname );
        exit (1);
      }

      if (!XAllocColor( display, cmap, &def ))
      { fprintf( stderr, "can't allocate color\n" );
        exit (1);
      }

      values.stipple = XCreateBitmapFromData( display,
                                              RootWindow( display, screen ),
                                              stipple[0], STIPPLE_W, STIPPLE_H);
      values.fill_style = FillSolid;
      values.foreground = def.pixel;
      values.background = WhitePixel( display, screen );
    }

    win->col[i] = XCreateGC( display,
                             RootWindow( display, screen ),
                             (GCForeground | GCBackground | GCStipple | GCFillStyle),
                             &values );

    XSetFont( display, win->col[i], win->font->fid );
  }

  values.function   = GXxor;
  values.plane_mask = 1;
  values.foreground = 1;
  values.background = 0;
  win->col[INVERT]  = XCreateGC( display,
                                 RootWindow( display, screen ),
                                 (GCForeground | GCBackground | GCFunction | GCPlaneMask),
                                 &values );

  /* display window */

  XMapWindow( display, win->win );

  /* wait until window appears */

  XSelectInput( display, win->win, ExposureMask );
  XWindowEvent( display, win->win, ExposureMask, &report );
}

/*---------------------------------------------------------------------------*/
